"""acts.core.model package."""

from __future__ import annotations

import numpy as np
import pandas as pd

from acts.core.model import util
from acts.core.model.base import MultinomialLogisticRegression
from acts.core.model.base import LogisticRegression

from acts.core.dataset.survey import divide


def TravelDecisionMLogit(df: pd.DataFrame, threshold: float = 0.05):
    """Travel Generation/Travel Decision MLogit Model.

    Returns:
        significant_vars: A list of column names representing the
            significant variables in the input DataFrame.
        output_df: Modified DataFrame containing all the rows where
            an agent went to travel (TRAVEL value == 1).
    """
    # Filter only travelling datasets
    output_filter = df["travel"] == 1
    output_df = df.copy(deep=True)[output_filter]
    output_df = output_df.drop(["travel"], axis=1)

    if output_filter.all():
        travel_significant_vars = [c for c in df.columns if c != "travel"]
        return travel_significant_vars, output_df

    mlogit, *_ = MLogit(df, "travel")
    return mlogit.get_significant_vars(threshold=threshold), output_df


def ActivityChoiceMLogit(df: pd.DataFrame, threshold: float = 0.05):
    """Activity Choice MLogit Model.

    Returns:
        significant_vars: A list of column names representing the
            significant variables in the input DataFrame.
        df: Return input dataframe.
    """
    mlogit = MLogit(df, "act")
    return mlogit.get_significant_vars(threshold=threshold), {
        choice: df[df["act"] == choice] for choice in df["act"].unique()
    }


def DestinationChoiceMLogit(dfs: list[pd.DataFrame], threshold: float = 0.05):
    """Destination Choice MLogit Model.

    Returns:
        significant_vars: A list of column names representing the
            significant variables in the input DataFrame.
        dfs: Return input dataframe.
    """
    output = {}
    for choice, df in dfs.items():
        # mlogit = MLogit(df, "destfm")
        print(df)
        # output[choice] = (
        #     mlogit.get_significant_vars(threshold=threshold),
        #     df,
        # )

    return output


def MLogit(
    df: pd.DataFrame,
    indep_var: str,
    *dep_vars: tuple(str),
    significant_vars: list[str] | None = None,
) -> pd.DataFrame:
    x, y = divide(df, indep_var=indep_var, dep_vars=list(dep_vars))

    try:
        mlogit_model = MultinomialLogisticRegression(y, x, fit_intercept=True)
        return mlogit_model.fit()

    except np.linalg.LinAlgError:
        logit_model = LogisticRegression(y, x, fit_intercept=True)
        return logit_model.fit()


def Logit(
    df: pd.DataFrame,
    indep_var: str,
    *dep_vars: tuple(str),
) -> pd.DataFrame:
    x, y = divide(df, indep_var=indep_var, dep_vars=list(dep_vars))

    logit_model = LogisticRegression(y, x, fit_intercept=True)
    return logit_model.fit()



"""Module containing generic utility functions."""

from __future__ import annotations

import uuid as uuidlib


__all__ = [
    # Constant exports
    "ACTS_NAMESPACE",
    # Function exports
    "generate_random_id",
    "generate_id",
]


ACTS_NAMESPACE = uuidlib.UUID("bd795e2b-6f74-4266-8d5a-40429e5fe7cb")


def generate_random_id(*, hex: bool = False) -> str:
    uuid = uuidlib.uuid4()
    if hex:
        return uuid.hex
    return str(uuid)


def generate_id(input_string: str, *, hex: bool = False) -> str:
    uuid = uuidlib.uuid5(ACTS_NAMESPACE, input_string)
    if hex:
        return uuid.hex
    return str(uuid)



"""Module containing OSM dataset loading functions."""

from __future__ import annotations

import pathlib
import re
import unicodedata

import pandas as pd

from acts.core import logging


__all__ = [
    # Function exports
    "load_osm",
]

logger = logging.get_logger(__name__)


def load_osm(name: str) -> pd.DataFrame:
    """Loads a parquet file from our datasets."""
    filename = f"{_slugify(name)}.parquet"
    filedir = pathlib.Path(__file__).parent.resolve()
    filepath = filedir.joinpath("parquets", filename)

    logger.running(f"Fetching {filepath!r}")
    output = pd.read_parquet(filepath)
    logger.success(f"Fetching {filepath!r}")

    return output


def _slugify(value: str, *, allow_unicode: bool = False) -> str:
    """Converts a string into a filename-safe version.

    Taken from github.com/django/django/blob/master/django/utils/text.py

    Convert to ASCII if `allow_unicode` is `False`. Convert spaces or
    repeated dashes to single dashes. Remove characters that aren't
    alphanumerics, underscores, or hyphens. Convert to lowercase. Also
    strip leading and trailing whitespace, dashes, and underscores.
    """
    value = str(value)
    if allow_unicode:
        value = unicodedata.normalize("NFKC", value)
    else:
        value = unicodedata.normalize("NFKD", value)
        value = value.encode("ascii", "ignore").decode("ascii")

    value = re.sub(r"[^\w\s-]", "", value.lower())
    return re.sub(r"[-\s]+", "-", value).strip("-_")



"""Module containing the utility functions for modeling."""

from __future__ import annotations

import pandas as pd

from acts.core.dataset.survey import divide
from acts.core.model.base import MultinomialLogisticRegression


def travel_choice_mlogit(df: pd.DataFrame) -> pd.DataFrame:
    return _base_choice_function(df, "ZONE")


def activity_choice_mlogit(df: pd.DataFrame) -> pd.DataFrame:
    return _base_choice_function(df, "ACTIVITY")


def destination_choice_mlogit(df: pd.DataFrame) -> pd.DataFrame:
    return _base_choice_function(df, "DES")


def mode_choice_mlogit(df: pd.DataFrame, *, verbose: bool = False) -> pd.DataFrame:
    return _base_choice_function(
        df,
        "MODEFM",
        dep_vars=[
            "DESTFM",
            "ACT",
            "TRVFREQ",
            "AGEB",
            "MEMB",
            "OVEHB",
            "NVEHB",
            "OCCUB",
            "MINCOMEB",
            "TRAVTIMEFM",
        ],
        verbose=verbose,
    )


def _base_choice_function(
    df: pd.DataFrame,
    indep_var: str,
    *,
    verbose: bool = False,
    dep_vars: list[str] | None = None,
) -> pd.DataFrame:
    x, y = divide(df, indep_var=indep_var, dep_vars=dep_vars)

    mlogit_model = MultinomialLogisticRegression(y, x, fit_intercept=True)
    mlogit_result = mlogit_model.fit()

    if verbose:
        print(mlogit_result.summary())

    return mlogit_result



"""Module containing the base model for the whole project."""

from __future__ import annotations

from statsmodels.api import add_constant
from statsmodels.base.model import LikelihoodModel
from statsmodels.discrete.discrete_model import (
    BinaryResultsWrapper as _BinaryResultsWrapper,
)
from statsmodels.discrete.discrete_model import BinaryModel
from statsmodels.discrete.discrete_model import Logit
from statsmodels.discrete.discrete_model import LogitResults
from statsmodels.discrete.discrete_model import MNLogit
from statsmodels.discrete.discrete_model import MultinomialResults
from statsmodels.discrete.discrete_model import (
    MultinomialResultsWrapper as _MultinomialResultsWrapper,
)
import numpy as np


class MultinomialLogisticRegression(MNLogit):
    def __init__(
        self,
        endog,
        exog,
        check_rank=True,
        fit_intercept=False,
        **kwargs,
    ):
        if fit_intercept:
            exog = add_constant(exog)
        super().__init__(endog, exog, check_rank=True, **kwargs)

    def fit(
        self,
        start_params=None,
        method="ncg",
        maxiter=35,
        full_output=True,
        disp=True,
        callback=None,
        **kwargs,
    ):
        if start_params is None:
            start_params = np.zeros((self.K * (self.J - 1)))
        else:
            start_params = np.asarray(start_params)

        def callback(x):  # Placeholder until check_perfect_pred
            return None

        # Skip calling super to handle results from LikelihoodModel
        mnfit = LikelihoodModel.fit(
            self,
            start_params=start_params,
            method=method,
            maxiter=maxiter,
            full_output=full_output,
            disp=disp,
            callback=callback,
            **kwargs,
        )

        mnfit.params = mnfit.params.reshape(self.K, -1, order="F")
        mnfit = MultinomialResults(self, mnfit)
        return MultinomialResultsWrapper(mnfit)


class MultinomialResultsWrapper(_MultinomialResultsWrapper):
    def get_significant_vars(self, threshold: float):
        """Return all the variables that has less than 5% p-value."""
        pvalues = self.pvalues.applymap(lambda x: True if x < threshold else False)
        output = {}
        for column in pvalues.columns:
            output[column] = list(pvalues[pvalues[column]].index)
        return output


class LogisticRegression(Logit):
    def __init__(
        self,
        endog,
        exog,
        check_rank=True,
        fit_intercept=False,
        **kwargs,
    ):
        if fit_intercept:
            exog = add_constant(exog)
        super().__init__(endog, exog, check_rank=True, **kwargs)

    def fit(
        self,
        start_params=None,
        method="ncg",
        maxiter=35,
        full_output=True,
        disp=True,
        callback=None,
        **kwargs,
    ):

        bnryfit = BinaryModel.fit(
            self,
            start_params=start_params,
            method=method,
            maxiter=maxiter,
            full_output=full_output,
            disp=disp,
            callback=callback,
            **kwargs,
        )

        discretefit = LogitResults(self, bnryfit)
        return BinaryResultsWrapper(discretefit)


class BinaryResultsWrapper(_BinaryResultsWrapper):
    def get_significant_vars(self, threshold: float):
        """Return all the variables that has less than 5% p-value."""
        pvalues = self.pvalues.applymap(lambda x: True if x < threshold else False)
        output = {}
        for column in pvalues.columns:
            output[column] = list(pvalues[pvalues[column]].index)
        return output



"""acts.core.dataset package."""

from __future__ import annotations

from typing import Callable

import pandas as pd

from acts.core.dataset.osm import load_osm as OSM
from acts.core.dataset.survey import divide


# fmt: off
__all__ = [
    # Function exports
    "load",
    "divide",

    # Constant exports
    "OSM",
]
# fmt: on


def load(
    nameorfile: str,
    *,
    dtype: Callable = OSM,
    **kwargs,
) -> pd.DataFrame:
    if dtype not in [OSM]:
        raise ValueError(f"Unknown dataset type: {dtype!r}")
    return dtype(nameorfile, **kwargs)



"""Module containing a custom JSON logger."""

from __future__ import annotations

from datetime import datetime
from datetime import timezone
import logging as _logging
import os

from pythonjsonlogger import jsonlogger as json_logging


# fmt: off
__all__ = [
    # Constant exports
    "CRITICAL",
    "DEBUG",
    "ERROR",
    "FATAL",
    "INFO",
    "WARN",
    "WARNING",

    # Variable exports
    "Logger",

    # Function exports
    "get_logger",
]
# fmt: on


# Logging levels, standard for logging extensions
DEBUG = _logging.DEBUG
INFO = _logging.INFO
WARNING = _logging.WARNING
WARN = _logging.WARN
ERROR = _logging.ERROR
CRITICAL = _logging.CRITICAL
FATAL = _logging.FATAL

# Standard Logger reference
Logger = _logging.Logger


class _SuccessAndRunningLogger(_logging.Logger):
    def success(self, *args, **kwargs):
        self._success_running("Success", *args, **kwargs)

    def running(self, *args, **kwargs):
        self._success_running("Running", *args, **kwargs)

    def _success_running(self, status: str, *args, **kwargs) -> None:
        extra = kwargs.pop("extra", {})
        if not extra:
            extra = {}
        new_extra = {"status": status}
        new_extra.update(extra)
        self.info(*args, extra=new_extra, **kwargs)


class _JSONFormatter(json_logging.JsonFormatter):
    def add_fields(
        self,
        log_record: dict,
        record: _logging.LogRecord,
        message_dict: dict,
    ):
        log_record["level"] = log_record.get("level", record.levelname).upper()
        log_record["timestamp"] = datetime.fromtimestamp(
            record.created,
            tz=timezone.utc,
        )

        super().add_fields(log_record, record, message_dict)


def get_logger(name: str | None = None, level: int = _logging.INFO, **kwargs):
    """Returns a logger object with additional JSON logging."""

    # Create the logger and set the wanted logging level
    _logging.setLoggerClass(_SuccessAndRunningLogger)
    logger = _logging.getLogger(name or "acts-core")
    logger.setLevel(level)

    # Create the JSON formatter instance
    logger_formatter = _JSONFormatter(
        json_indent=_get_json_indent(),
        timestamp=False,
    )

    # Check if we already have a handler (for AWS lambda really)
    if len(logger.handlers) != 0:
        logger.handlers[0].setFormatter(logger_formatter)
    else:
        logger_handler = _logging.StreamHandler()
        logger_handler.setFormatter(logger_formatter)
        logger.addHandler(logger_handler)

    return logger


def _get_json_indent() -> int | None:
    """Returns indent from environment variable `TWS_JSON_LOG_INDENT`."""
    tws_json_log_indent = os.environ.get("TWS_JSON_LOG_INDENT")

    try:
        return int(tws_json_log_indent) or None

    except Exception:
        # We can pass here because we want the function to
        # proceed if `TWS_JSON_LOG_INDENT` is not a valid int
        pass

    return None



"""ACTS Core is the Core Backend Python Library for Project ACTS."""

from __future__ import annotations

import setuptools
import os


PACKAGE_ROOT_DIR = os.path.abspath(os.path.dirname(__file__))
DOCLINES = __doc__.split("\n")


def get_requirements() -> list[str]:
    with open("requirements/base.txt", "r") as fp:
        package_list = fp.readlines()
        package_list = [package.rstrip() for package in package_list]

    return package_list


def get_version() -> str:
    version = {}
    version_file = os.path.join(
        PACKAGE_ROOT_DIR,
        "acts",
        "core",
        "version.py",
    )
    with open(os.path.join(version_file)) as fp:
        exec(fp.read(), version)

    return version["__version__"]


setuptools.setup(
    name="acts.core",
    description=DOCLINES[0],
    author="Project ACTS - Software Development Team",
    version=get_version(),
    install_requires=get_requirements(),
    include_package_data=True,
    packages=setuptools.find_namespace_packages(
        include=[
            "acts.*",
        ]
    ),
    python_requires=">=3.7",
)



"""Module containing survey-related data loading functions."""

from __future__ import annotations

import pandas as pd


__all__ = [
    # Function exports
    "divide",
]


def divide(df: str, **kwargs) -> pd.DataFrame:
    """Load and return the sample coded dataset (classification)."""

    # Normalize values
    if kwargs.get("normalize", False):
        df = (df - df.min()) / (df.max() - df.min())

    # Convert all columns to lowercase
    df.columns = map(str.lower, df.columns)

    independent_column = kwargs.get("indep_var", "mode").lower()

    dependent_colums = list(df.columns)
    dependent_colums = kwargs.get("dep_vars", dependent_colums) or dependent_colums
    dependent_colums = [c.lower() for c in dependent_colums]

    if independent_column in dependent_colums:
        dependent_colums.remove(independent_column)

    x = df[dependent_colums]
    y = df[independent_column]

    return x, y
